#include "gui.h"
#include "qdebug.h"
#include "ui_gui.h"

#include "gui/fancyslider.h"

#include <QDateTime>
#include <QScreen>

Gui::Gui(QWidget *parent) :
    QMainWindow(parent),
    ui(new Ui::Gui)
{
    ui->setupUi(this);

    QScreen *screen = QGuiApplication::primaryScreen();
    QRect  screenGeometry = screen->geometry();
    setGeometry(screenGeometry.right() - width(), screenGeometry.bottom() - height(), width(), height());

    file = new QFile("efs.log");
    if (file->open(QIODevice::Truncate | QIODevice::WriteOnly))
    {
        stream = new QTextStream(file);
    }
    else
    {
        throw new std::runtime_error("Failed to open log file.");
    }

    logLine("Starting up!");

    client = new Client("ws://127.0.0.1", 12345);
    client->connect(callbackFunctionPtr);
    client->requestDeviceList();

    logLine("Client started!");

    logLine("Started OK!");
    setWindowFlag(Qt::WindowStaysOnTopHint, true);
//    setWindowFlags(Qt::CustomizeWindowHint | Qt::WindowTitleHint | Qt::WindowMinMaxButtonsHint |  );

    modules.append(&speedMod);
    connect(ui->speedEnabledBox, &QCheckBox::stateChanged, &speedMod, &SpeedModule::setEnabled);
    connect(&speedMod, &SpeedModule::EnabledChanged, ui->speedEnabledBox, &QCheckBox::setChecked);
    connect(ui->speedTopSlider, &FancySlider::valueChanged, &speedMod, &SpeedModule::setTopSpeed);
    connect(&speedMod, &SpeedModule::TopSpeedChanged, ui->speedTopSlider, &FancySlider::setValue);
    ui->speedTopSlider->setMax(100.0);
    connect(ui->speedBumpFreqSlider, &FancySlider::valueChanged, &speedMod, &SpeedModule::setWaveFrequency);
    connect(ui->speedBumpFreqSlider, &FancySlider::valueChanged, this, [](double v){qDebug() << "Wave freq " << v;});
    connect(&speedMod, &SpeedModule::WaveFrequencyChanged, ui->speedBumpFreqSlider, &FancySlider::setValue);
    ui->speedBumpFreqSlider->setMax(5.0);
    ui->speedBumpFreqSlider->setMin(0.1);
    connect(ui->speedBumpAmpSlider, &FancySlider::valueChanged, &speedMod, &SpeedModule::setWaveAmplitude);
    connect(&speedMod, &SpeedModule::WaveAmplitudeChanged, ui->speedBumpAmpSlider, &FancySlider::setValue);
    ui->speedBumpAmpSlider->setMax(0.5);
    connect(ui->speedOutputPowerSlider, &FancySlider::valueChanged, &speedMod, &SpeedModule::setMaxOutput);
    connect(&speedMod, &SpeedModule::MaxOutputChanged, ui->speedOutputPowerSlider, &FancySlider::setValue);
    ui->speedOutputPowerSlider->setMax(1.0);

    modules.append(&rpmMod);
    connect(ui->rpmEnabledBox, &QCheckBox::stateChanged, &rpmMod, &RPMModule::setEnabled);
    connect(&rpmMod, &RPMModule::EnabledChanged, ui->rpmEnabledBox, &QCheckBox::setChecked);
    connect(ui->rpmTopSlider, &FancySlider::valueChanged, &rpmMod, &RPMModule::setTopRPM);
    connect(&rpmMod, &RPMModule::TopRPMChanged, ui->rpmTopSlider, &FancySlider::setValue);
    ui->rpmTopSlider->setMax(3000.0);
    connect(ui->rpmBottomSlider, &FancySlider::valueChanged, &rpmMod, &RPMModule::setBottomRPM);
    connect(&rpmMod, &RPMModule::BottomRPMChanged, ui->rpmBottomSlider, &FancySlider::setValue);
    ui->rpmBottomSlider->setMax(1000.0);
    connect(ui->rpmOutputPowerSlider, &FancySlider::valueChanged, &rpmMod, &RPMModule::setMaxOutput);
    connect(&rpmMod, &RPMModule::MaxOutputChanged, ui->rpmOutputPowerSlider, &FancySlider::setValue);
    ui->rpmOutputPowerSlider->setMax(1.0);

    modules.append(&crimeMod);
    connect(ui->crimeEnableBox, &QCheckBox::stateChanged, &crimeMod, &CrimeModule::setEnabled);
    connect(&crimeMod, &CrimeModule::EnabledChanged, ui->crimeEnableBox, &QCheckBox::setChecked);
    connect(ui->crimeRateSlider, &FancySlider::valueChanged, &crimeMod, &CrimeModule::setSecondsPerEuro);
    connect(&crimeMod, &CrimeModule::SecondsPerEuroChanged, ui->crimeRateSlider, &FancySlider::setValue);
    ui->crimeRateSlider->setMax(100.0);
    connect(ui->crimePowerMultiplier, &FancySlider::valueChanged, &crimeMod, &CrimeModule::setPowerMultiplier);
    connect(&crimeMod, &CrimeModule::PowerMultiplierChanged, ui->crimePowerMultiplier, &FancySlider::setValue);
    ui->crimePowerMultiplier->setMax(2.0);

    modules.append(&jobMod);
    connect(ui->jobRewardEnableBox, &QCheckBox::stateChanged, &jobMod, &JobModule::setEnabled);
    connect(&jobMod, &JobModule::EnabledChanged, ui->jobRewardEnableBox, &QCheckBox::setChecked);
    connect(ui->jobRewardMoneyRateSlider, &FancySlider::valueChanged, &jobMod, &JobModule::setSecondsPerEuro);
    connect(&jobMod, &JobModule::SecondsPerEuroChanged, ui->jobRewardMoneyRateSlider, &FancySlider::setValue);
    ui->jobRewardMoneyRateSlider->setMax(10);
    connect(ui->jobRewardXPRateSlider, &FancySlider::valueChanged, &jobMod, &JobModule::setSecondsPerXP);
    connect(&jobMod, &JobModule::SecondsPerXPChanged, ui->jobRewardXPRateSlider, &FancySlider::setValue);
    ui->jobRewardXPRateSlider->setMax(30);
    connect(ui->jobRewardPowerBoostSlider, &FancySlider::valueChanged, &jobMod, &JobModule::setPowerBoost);
    connect(&jobMod, &JobModule::PowerBoostChanged, ui->jobRewardPowerBoostSlider, &FancySlider::setValue);
    ui->jobRewardPowerBoostSlider->setMax(1.0 );

    modules.append(&distMod);
    connect(ui->distanceRewardEnabledBox, &QCheckBox::stateChanged, &distMod, &JobModule::setEnabled);
    connect(&distMod, &JobModule::EnabledChanged, ui->distanceRewardEnabledBox, &QCheckBox::setChecked);
    connect(ui->distanceRewardDistSlider, &FancySlider::valueChanged, &distMod, &DistanceModule::setDistance);
    connect(&distMod, &DistanceModule::DistanceChanged, ui->distanceRewardDistSlider, &FancySlider::setValue);
    ui->distanceRewardDistSlider->setMax(50);
    connect(ui->distanceRewardRandSlider, &FancySlider::valueChanged, &distMod, &DistanceModule::setRandomness);
    connect(&distMod, &DistanceModule::RandomnessChanged, ui->distanceRewardRandSlider, &FancySlider::setValue);
    ui->distanceRewardRandSlider->setMax(1);
    connect(ui->distanceRewardDurationSlider, &FancySlider::valueChanged, &distMod, &DistanceModule::setDuration);
    connect(&distMod, &DistanceModule::DurationChanged, ui->distanceRewardDurationSlider, &FancySlider::setValue);
    ui->distanceRewardDurationSlider->setMax(60.0);
    connect(ui->distanceRewardPowerBoostSlider, &FancySlider::valueChanged, &distMod, &DistanceModule::setPowerBoost);
    connect(&distMod, &DistanceModule::PowerBoostChanged, ui->distanceRewardPowerBoostSlider, &FancySlider::setValue);
    ui->distanceRewardPowerBoostSlider->setMax(1.0);

    loadSettings();

    for (RuleModule* module: modules)
    {
        connect(module, &RuleModule::statusUpdate, this, &Gui::rulesStatusUpdate);
    }

    ui->tabWidget->setCurrentIndex(0);

    timer = new QTimer(this);
    connect(timer, &QTimer::timeout, this, &Gui::update);
    timer->start(UPDATE_INTERVAL_MS);
}

Gui::~Gui()
{
    delete ui;
}

//void Gui::doubleConnect(QObject* objectA, QObject* objectB, const QMetaMethod aSignal, const QMetaMethod aSlot, const QMetaMethod bSignal, const QMetaMethod bSlot )
//{
//    connect(objectA, aSignal, objectB, bSlot);
//    connect(objectB, bSignal, objectA, aSlot);
//}

void Gui::loadSettings()
{

    QSettings settings;
    qDebug() << "Loading settings from " << settings.fileName();
    for (RuleModule* module : modules)
    {
        module->load(settings);
    }
}

void Gui::saveSettings()
{
    QSettings settings;
    for (RuleModule* module : modules)
    {
        module->save(settings);
    }

}

void Gui::shutdown()
{
    logLine("Shutting down!");
    emit shuttingDown();
    if (client->clientConnected)
    {
        logLine("Disconnecting from client!");
        client->stopAllDevices();
        logLine("Devices stopped.");
        client->disconnect();
        logLine("Client disconnected.");
    }

    stream->flush();
    file->close();

}

void Gui::closeEvent(QCloseEvent* event)
{
//    event->ignore();
    saveSettings();
    shutdown();
    QApplication::quit();
    event->accept();
}

void Gui::updateStatus()
{
    QString statusHTML = "<html><head/><body>";
    QString connectionStatus = "<p>";
    QString hintText = "";
    if (client->clientConnected)
    {
        connectionStatus += "Connected to Intiface!";
        if (client->getDevices().size() == 0)
        {
            connectionStatus += " No devices available.";
            hintText += "Put your device in pairing mode, then press 'Devices / Start Scanning' in Intiface Central.";
        }
    }
    else
    {
        connectionStatus += "Can't connect to Intiface!";
        hintText += "Make sure <a href=\"https://intiface.com/central/\">Intiface Central</a> is installed and running.<br/>If Intiface is on another device, enter the address (ws://...) on the 'Config' tab.";
    }
    connectionStatus += "</p>";

    statusHTML += connectionStatus;
    if (!hintText.isEmpty())
    {
        statusHTML += "<p><em>" + hintText + "</em></p>";
    }

    if (!rulesStatusText.isEmpty())
    {
        statusHTML += "<p>" + rulesStatusText + "</p>";
    }

    statusHTML += "</body></html>";
    ui->statusLabel->setText(statusHTML);
}

void Gui::update()
{
    QMutexLocker lk(&telemetryMutex);

    long time = latestTelemetry.timestamp;
    if (time == 0)
    {
        time = QDateTime::currentMSecsSinceEpoch() * 1000;
    }


    long deltaUS = lastTime > 0 ? time - lastTime : 0;
    double delta = deltaUS / (1000.0 * 1000.0);

    double deltaDistance = latestTelemetry.speed * delta;
    distanceCounter += deltaDistance;

    lastTime = time;

    TelemetryState tele;
    if (!paused)
        tele = latestTelemetry;

    tele.distance = distanceCounter;

    double power = 0.0;

    for (RuleModule* module: modules)
    {
        power = module->process(tele, power, delta);
    }

    power = std::clamp(power, 0.0, 1.0);

    if (client->clientConnected && !paused)
    {

        std::vector<DeviceClass> devices = client->getDevices();
        for (DeviceClass device : devices)
        {
            logLine(QString("Power to %2: %1").arg(power).arg(QString::fromStdString(device.deviceName)));
            client->sendScalar(device, power);
        }
    }

    updateStatus();

    ui->distanceLabel->setText(QString("%1").arg(tele.distance / 1000.0, 0, 'f', 3));
    ui->outputGraph->addPoint(power);
}

void Gui::setPause(bool pause)
{
    paused = pause;
    if (paused)
        logLine("Game paused.");
    else
        logLine("Game unpaused.");
}

void Gui::logLine(QString line)
{
    qDebug() << line;
    //    *stream << line << "\n";
    //    stream->flush();
}

void Gui::receiveTelemetry(TelemetryState tele)
{
    if (!paused)
    {
        //        logLine(QString("speed %1 rpm %2").arg(tele.speed).arg(tele.rpm));
        QMutexLocker lk(&telemetryMutex);

        latestTelemetry = tele;

        ui->speedLabel->setText(QString("%1").arg(tele.speed * 3.6, 4, 'f', 1));
        ui->rpmLabel->setText(QString("%1").arg(tele.rpm, 0, 'f', 0));
    }
}

void Gui::rulesStatusUpdate(QString status)
{
    rulesStatusText = status;
    QTimer::singleShot(1000, [this](){ rulesStatusText = ""; });
}

void Gui::gameplayEvent(QString event, QMap<QString, QVariant> attributes)
{
    qDebug() << "Game event: " << event << ", atts: " << attributes;
    for (RuleModule* module: modules)
    {
        module->gameEvent(event, attributes);
    }
}

void Gui::callbackFunction(const mhl::Messages msg)
{
    if (msg.messageType == mhl::MessageTypes::DeviceList)
    {
        logLine("Device List callback");
    }
    if (msg.messageType == mhl::MessageTypes::DeviceAdded)
    {
        logLine("Device Added callback");
    }
    if (msg.messageType == mhl::MessageTypes::ServerInfo)
    {
        logLine("Server Info callback");
    }
    if (msg.messageType == mhl::MessageTypes::DeviceRemoved)
    {
        logLine("Device Removed callback");
    }
    if (msg.messageType == mhl::MessageTypes::SensorReading)
    {
        logLine("Sensor Reading callback");
    }
}

void Gui::on_jobCompleteTestButton_pressed()
{
    QMap<QString, QVariant> atts;
    atts["earned.xp"] = QVariant::fromValue(125);
    atts["revenue"] = QVariant::fromValue(2041);
    gameplayEvent("job.delivered", atts);
}

