#ifndef RULEMODULE_H
#define RULEMODULE_H

#include <QObject>
#include <QSettings>

#include "telemetry.h"



// I'm sorry I'm just addicted to efficiency ok
#define FANCY_PROPERTY(Type,Name,Default) \
Q_PROPERTY(Type Name READ get##Name WRITE set##Name NOTIFY Name##Changed); \
protected: \
    Type Name = Default; \
public: \
    Type get##Name() { return Name; } \
    Q_SIGNAL void Name##Changed(Type newValue); \
public slots: \
    void set##Name(Type value) { bool changed = (Name != value); Name = value; if (changed) emit Name##Changed(Name); };

class RuleModule : public QObject
{
    Q_OBJECT;

public:
    explicit RuleModule(QObject *parent = nullptr);

    virtual double process(const TelemetryState& telemetry, double previous, double delta) = 0;
    virtual void gameEvent(const QString name, const QMap<QString, QVariant> attributes) {};

    void load(QSettings& settings);
    void save(QSettings& settings) const;

    FANCY_PROPERTY(bool, Enabled, false);

signals:
    void statusUpdate(QString status);

public slots:
    virtual void gameEvent(QString eventType, double param) {}

};

class SpeedModule: public RuleModule
{
    Q_OBJECT;
public:
    SpeedModule() { setObjectName("speed"); }
    virtual double process(const TelemetryState& telemetry, double previous, double delta) override;

    FANCY_PROPERTY(double, TopSpeed, 65.0);
    FANCY_PROPERTY(double, MaxOutput, 1.0);
    FANCY_PROPERTY(double, WaveFrequency, 3.0);
    FANCY_PROPERTY(double, WaveAmplitude, 0.25);

private:
    double sine_x = 0.0;
};

class RPMModule: public RuleModule
{
    Q_OBJECT;
public:
    RPMModule() {setObjectName("rpm");}
    virtual double process(const TelemetryState& telemetry, double previous, double delta) override;

    FANCY_PROPERTY(double, TopRPM, 2500.0);
    FANCY_PROPERTY(double, BottomRPM, 500.0);
    FANCY_PROPERTY(double, MaxOutput, 1.0);
};

class CrimeModule: public RuleModule
{
    Q_OBJECT;
public:
    CrimeModule() {setObjectName("crime");}
    virtual double process(const TelemetryState& telemetry, double previous, double delta) override;
    virtual void gameEvent(const QString name, const QMap<QString, QVariant> attributes) override;

    FANCY_PROPERTY(double, SecondsPerEuro, 30.0 / 400.0);
    FANCY_PROPERTY(double, PowerMultiplier, 0.1);

private:
    double timeout = 0.0;
    QString whatYouDid = "";
};

class JobModule: public RuleModule
{
    Q_OBJECT;
public:
    JobModule() {setObjectName("job");}
    virtual double process(const TelemetryState& telemetry, double previous, double delta) override;
    virtual void gameEvent(const QString name, const QMap<QString, QVariant> attributes) override;

    FANCY_PROPERTY(double, SecondsPerEuro, 30.0 / 7000.0);
    FANCY_PROPERTY(double, SecondsPerXP, 30.0 / 500.0);
    FANCY_PROPERTY(double, PowerBoost, 0.7);

private:
    double totalTime = 0.0;
    double timeout = 0.0;
};

class DistanceModule: public RuleModule
{
    Q_OBJECT;
public:
    DistanceModule() {setObjectName("distance");}
    virtual double process(const TelemetryState& telemetry, double previous, double delta) override;

    FANCY_PROPERTY(double, Distance, 10.0);
    FANCY_PROPERTY(double, Randomness, 1.0);
    FANCY_PROPERTY(double, Duration, 30.0);
    FANCY_PROPERTY(double, PowerBoost, 0.7);

private:
    double nextReward = 0.0;
    double timeout = 0.0;

};

#endif // RULEMODULE_H
